#include <mda/common/audio.h>

#include "SSSDThread.h"
#include "TAudioShared.h"

extern "C" {
#include "sssd.h"
#include "internals.h"
}

TInt CSSSDThread::ThreadFunction(TAny *aData) {
    TAudioShared& shared = *((TAudioShared*)aData);

    // signal to the client we are alive
    // signalled off when destroying this thread
    shared.iAliveSem.Wait();

    CSSSDThread *sssdThread = CSSSDThread::Create(aData);
    if (sssdThread == NULL) {
        return KErrGeneral;
    }

    // active scheduler has been constructed
    // will run in wait loop until it's time to end the thread
    CActiveScheduler::Start();

    // active schedule has been stopped, so delete the thread object
    delete sssdThread;

    // tell owning thread it's safe to exit
    shared.iAliveSem.Signal();

    return KErrNone;
}

CSSSDThread *CSSSDThread::Create(TAny *aData) {
    CSSSDThread *self = new CSSSDThread(aData);
    if (self == NULL) return self;

    if (self->Construct() != KErrNone) {
        delete self;
        return NULL;
    }

    return self;
}

TInt CSSSDThread::Construct() {
    // create cleanup stack
    iCleanupStack = CTrapCleanup::New();
    if (iCleanupStack == NULL) {
        return KErrNoMemory;
    }

    TInt err = KErrNone;
    TRAP(err, ConstructL());

    return err;
}

void CSSSDThread::ConstructL() {
    // create active scheduler
    iActiveScheduler = new (ELeave) CActiveScheduler;
    CActiveScheduler::Install(iActiveScheduler);

    Dll::SetTls(this);
    RThread().SetExceptionHandler(ExcHandler, 0xffffffff);
}

CSSSDThread::CSSSDThread(TAny *aData) : iShared(*((TAudioShared*)aData)) {
    iSSSDAudio = iShared.iSSSDAudio;
    iBufferPtr.Set(TPtrC8((TUint8 *)iSSSDAudio->playBuffer, iSSSDAudio->audioBufferSize));
    switch (iSSSDAudio->channels) {
        case 1: {
            iSet.iChannels = TMdaAudioDataSettings::EChannelsMono;
            break;
        }
        default: {
            iSet.iChannels = TMdaAudioDataSettings::EChannelsStereo;
            break;
        }
    }
    switch (iSSSDAudio->freq) {
        case 8000: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate8000Hz;
            break;
        }
        case 11025: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate11025Hz;
            break;
        }
        case 16000: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate16000Hz;
            break;
        }
        case 22050: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate22050Hz;
            break;
        }
        case 32000: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate32000Hz;
            break;
        }
        case 44100: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate44100Hz;
            break;
        }
        case 48000: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate48000Hz;
            break;
        }
        default: {
            iSet.iSampleRate = TMdaAudioDataSettings::ESampleRate44100Hz;
            break;
        }
    }
}

CSSSDThread::~CSSSDThread() {
    delete iStream;
    delete iActiveScheduler;
    delete iCleanupStack;
}


void CSSSDThread::ExcHandler(TExcType aExc) {
    // exception handler entry function
    CSSSDThread *threadPointer = (CSSSDThread*)Dll::Tls();
    threadPointer->HandleException(aExc);
}

void CSSSDThread::HandleException(TExcType aExc) {
    // handle exceptions
    // EExcUserInterrupt is reserved for inter-thread communication
    switch(aExc) {
        case EExcUserInterrupt: {
            // perform the client request
			switch( iShared.iCmd ) {
				case ECmdStartMixer: {
					TRAPD(err,StartMixerL());
					if (err != KErrNone) {
					    iError = err;
					}
					break;
				}
				case ECmdStopMixer: {
					StopMixer();
					break;
				}
				case ECmdDestroyMixer: {
					CActiveScheduler::Stop();
					break;
				}

			}
            break;
        }
        default: {
            // unknown exception - exit this thread
            CActiveScheduler::Stop();
            break;
        }
    }
}

void CSSSDThread::StartMixerL() {
	iStream = CMdaAudioOutputStream::NewL( *this );
	iStreamOpen = EFalse;
	iStream->Open( &iSet );
}


void CSSSDThread::StopMixer() {
    if (iStream != NULL) {
    	if (iStreamOpen) iStream->Stop();
    	delete iStream;
    	iStream = NULL;
    }
}

void CSSSDThread::FillBufferL() {
    unsigned short *playbuffer;
    struct ga_playing* curplaying;
    short leftsample;
    unsigned short *curposleft;
    unsigned char *tmpsp;
    int i, j;
    int local_num_playing, abufsize;
    _REGISTER int sampAccumL;
    int fillBackground=1;

    int CBRflag, blockUsec;

    struct sssdaudio *ga = iSSSDAudio;

    playbuffer=(unsigned short *)ga->playBuffer;
    abufsize=ga->audioBufferSize;
    CBRflag=ga->flags & SSSDSYS_CONSTRATE;
    blockUsec=((SLEEP_TYPE)abufsize*1000000)/ga->freq;

    /* Loop ... playing sound */
//playloop:

    LOCK_BUFFERS(ga);

//    if(ga->threadSignal==1) {
//	    ga->threadSignal=2;
//	    UNLOCK_BUFFERS(ga);
//	    EXIT_THREAD(ga);
//	}

    local_num_playing=ga->num_playing;

    memset((void *)playbuffer,0x00,abufsize);
    if(fillBackground==1) {
        if( ga->background_stream == NULL ) {
//	        memset((void *)playbuffer,0x80,abufsize);
	        memset((void *)playbuffer,0x00,abufsize);
	        fillBackground=0;
        } else {
            fillBackground=1;
	        /* Preset from background stream */
            if( fread(playbuffer, 1, abufsize, ga->background_stream)
                < abufsize )
            {
                /* Close */
                pclose( ga->background_stream );
                ga->background_stream = NULL;
                ga->background_stream_bytes = 0;
                /* Clear to zero after all */
//	            memset((void *)playbuffer,0x80,abufsize);
	            memset((void *)playbuffer,0x00,abufsize);
            } else {
            	/* Store stream bytes */
            	ga->background_stream_bytes += abufsize;
            }
        }
    }

    /* Needs mixing? */
    if( local_num_playing > 0 )
    {
        /* Create playbuffer */
        curposleft  = playbuffer;
	    fillBackground=1;

        for( i = abufsize>>1; i > 0; i-- )
        {
            curplaying = ga->playing;
//            sampAccumL=sample8to16(*curposleft)<<8;
            sampAccumL=(*curposleft)<<8;

            for( j = 0; j < local_num_playing; j++ )
            {
                /* Get source samples and advance sample positions */
                tmpsp=(unsigned char *)curplaying->curpos;
//                if(!((curplaying->voiceflags)&PLAYFLAG_CHANCHANGE)) {
                    leftsample=sample8to16(*tmpsp);
//                } else {
//                    int i=sample8to16(*tmpsp);
//                    tmpsp++;
//                    i+=sample8to16(*tmpsp);
//                    leftsample=i>>1;
//                }
                tmpsp++;
                curplaying->curpos=(void *)tmpsp;

                /* Mangle left and right components with the  */
                /* pan/volume coefficients                    */
                sampAccumL += (int)leftsample  * curplaying->b;

                /* Check if we are at the end */
                if( curplaying->repeat_count != 1 && curplaying->curpos == curplaying->loopend) {

                    curplaying->curpos = curplaying->looppos;
                    if( curplaying->repeat_count > 1 ) curplaying->repeat_count--;  /* sample doesn't loop forever */

		        } else
		        if( curplaying->repeat_count == 1 && curplaying->curpos == curplaying->playend) {
                    /* Done - delete entry from list */

                    ga->num_playing--;
			        local_num_playing--;

                    if( j != local_num_playing )
                        ga->playing[j] = ga->playing[local_num_playing];

                    /* Check entry again */
                    curplaying--;
                    j--;

		        }

                /* Next in playlist */
                curplaying++;
            }

		    sampAccumL>>=8;
		    if(sampAccumL>32767) sampAccumL=32767;
		    else
		    if(sampAccumL<-32768) sampAccumL=-32768;

//	        *curposleft=sample16to8(sampAccumL);
	        *curposleft=(sampAccumL);

            /* Advance pointers */
            curposleft++;

        }
    }

    UNLOCK_BUFFERS(ga);

    /* Write playbuffer to audio device */
    if(CBRflag || fillBackground) { /* Don't send empty buffer unless told to */
        // SEND_BUFFER_TO_DEVICE(playbuffer);
        iStream->WriteL(iBufferPtr);

    } else {
	    /* Instead- sleep 1 blocks worth of data -probably better do suspend and get a signal from play() */
        SLEEP_CALL(blockUsec);
    }

    // goto playloop;
}

void CSSSDThread::MaoscBufferCopied( TInt aError, const TDesC8& /*aBuffer*/ ) {
	if( aError ) {
		iError = aError;
    } else {
		TRAPD(err,FillBufferL());
		if (err != KErrNone) {
		    iError = err;
		}
	}
}

void CSSSDThread::MaoscOpenComplete( TInt aError ) {
	if( aError ) {
		iError = aError;
		iStreamOpen = EFalse;
	} else {
		iStreamOpen = ETrue;
		iStream->SetVolume( iStream->MaxVolume() );
		TRAPD(err,FillBufferL());
		if (err != KErrNone) {
		    iError = err;
		}
	}
}

void CSSSDThread::MaoscPlayComplete( TInt aError ) {
    if (aError == KErrUnderflow) {
		iStream->SetVolume( iStream->MaxVolume() );
		TRAPD(err,FillBufferL());
		if (err != KErrNone) {
		    iError = err;
		}
    } else {
		iError = aError;
		iStreamOpen = EFalse;
		delete iStream;
		iStream = NULL;
	}

}
