
#include "sssd.h"
#include "internals.h"

/*
    Start gameaudio system: open device, setup device
    parameters, initialize tables, start thread.  Returns -1 on error.
    stopAudio does not need to be called if startAudio does not succeed.
*/


/* Audio Management similar to Linux OSS/Free */
void WIN_AudioWrite(struct sssdaudio *ga, void *buf, int size) {

   struct WinPlatformData *pdata;
   LPWAVEHDR  hdr;
   void *buffer;
   int i1,i2,i3;

   pdata=(struct WinPlatformData *)ga->platformData;

   /* Get next header */
   i2=pdata->nextBuff;
   buffer=pdata->WinBuffer[i2];
   hdr=&pdata->headers[i2];
   i2++;i2%=ga->numAudioBuffers;

   /* need to wait for system to set WHDR_DONE bit */
   i1=pdata->sleepMS;
   i3=0;
   while((hdr->dwFlags&WHDR_DONE)==0) {SLEEP_CALL(i1);i3++;}

   /* Clear 'done' bit in header */
   hdr->dwFlags^=WHDR_DONE;

   /* Copy data to next buffer */
   memcpy(buffer,buf, size);

   /* Send to device */
   i1=waveOutWrite(pdata->hWaveOut, hdr, sizeof(WAVEHDR));
   if(i1!=MMSYSERR_NOERROR) printf("ERROR: waveOutWrite %d\n",i1);
   else {

	   pdata->nextBuff=i2;

   }
}

/* Include Mixer threads */

  #include "playthrd16s.c"
  #include "playthrd8s.c"
  #include "playthrd16m.c"
  #include "playthrd8m.c"

/* Proto rather than function for toollibs */
int Log2(int value) {
   int count=0;
   while(!(value&0x1)) {count++;value>>=1;}
return count;
}


int sssd_startAudio( struct sssdaudio* ga, const char* device_name )
{


	unsigned int ui1, i1;
	WAVEFORMATEX wfx;
    struct WinPlatformData *pdata=NULL;
	unsigned char *buffer;
    LPTHREAD_START_ROUTINE thread_rtn;

    #ifdef WIN_DEBUG
	printf("sssd_startAudio()\n");
    #endif

	/* Init variables */
    ga->paused      = 0;
    ga->num_samples = 0;
    ga->num_playing = 0;
    ga->playing_count = 0;
    ga->background_stream = NULL;
    ga->background_stream_bytes = 0;
	ga->buffer_mutex=NULL;
    ga->lastERR=0;
	ga->platformData=NULL;

    /* Set size of audio buffer */
    if(ga->audioBufferSize==0) {
       ga->audioBufferSize=256*ga->byps;
       if(ga->freq>22000) ga->audioBufferSize<<=1;
       if(ga->freq>44000) ga->audioBufferSize<<=1;
    }

    #ifdef WIN_DEBUG
	printf("\tallocate arrays\n");
    #endif

    /* Allocate arrays */
	ga->playing = 0;
	ga->samples = 0;
	if (ga->max_voices) {
		ga->samples=malloc(sizeof(struct ga_sample)*ga->max_samples);
		if(ga->samples==NULL) {
		ga->lastERR=ERR_SAMPLES_ALLOC;
		return ERR_SAMPLES_ALLOC;
		}
		ga->playing=malloc(sizeof(struct ga_playing)*ga->max_voices);
		if(ga->playing==NULL) {
		free(ga->samples);
		ga->lastERR=ERR_VOICES_ALLOC;
		return ERR_VOICES_ALLOC;
		}
		ga->playBuffer=malloc(ga->audioBufferSize);
		if(ga->playBuffer==NULL) {
		free(ga->samples);
		free(ga->playing);
		ga->lastERR=ERR_PLAYBUFFER_ALLOC;
		return ERR_PLAYBUFFER_ALLOC;
		}
	}

   	/* setup platform specific data and pointer */
    ga->platformData=malloc(sizeof(struct WinPlatformData));
	if(ga->platformData==NULL) {
		ga->lastERR=ERR_MEMORY_ALLOC;
		goto err;
	}
    pdata=(struct WinPlatformData *)ga->platformData;
	pdata->buffCount=0;
	pdata->nextBuff=0;
	ga->threadSignal=0;
	pdata->headers=NULL;
    pdata->sleepMS=(((ga->audioBufferSize/ga->byps)*1000000)/ga->freq)>>1;  /* Sleep(ms) is acutally Sleep(us)? */

	/* Setup audio device */

    wfx.wFormatTag=WAVE_FORMAT_PCM;
    wfx.nChannels=ga->channels;
    wfx.nSamplesPerSec=ga->freq;
    wfx.wBitsPerSample=ga->bits;
    wfx.nBlockAlign=ga->channels;if(ga->bits==16) wfx.nBlockAlign*=2;
	wfx.nAvgBytesPerSec=wfx.nBlockAlign*ga->freq;
    wfx.cbSize=0;

    #ifdef WIN_DEBUG
	printf("\tOpen audio device\n");
    #endif

	for(ui1=0; ui1<waveOutGetNumDevs(); ui1++) {
		if(waveOutOpen( &pdata->hWaveOut,
			WAVE_MAPPER,
			&wfx,
			(DWORD) NULL,
			(DWORD) pdata,
			CALLBACK_NULL) == MMSYSERR_NOERROR)
			break;
	}


	/* No suitable device found */
	if(ui1==waveOutGetNumDevs()) {
		ga->lastERR=ERR_DEVICE_OPEN;
		goto err;
	}


    #ifdef WIN_DEBUG
	printf("\tAlocate audio buffers\n");
    #endif

	/* Allocate memory for audio buffers */
    buffer=malloc((ga->audioBufferSize+sizeof(WAVEHDR))*ga->numAudioBuffers);
	if(buffer==NULL) {
		ga->lastERR=ERR_MEMORY_ALLOC;
		goto err;
	}

	pdata->WinBuffer=malloc(sizeof(void *)*ga->numAudioBuffers);
	if(buffer==NULL) {
		ga->lastERR=ERR_MEMORY_ALLOC;
		goto err;
	}

    #ifdef WIN_DEBUG
	printf("\tClear Headers (WHDR)\n");
    #endif

    /* clear audio headers */
	for(i1=0;i1<(sizeof(WAVEHDR)*ga->numAudioBuffers);i1++)
	   buffer[i1]=0x00;


	pdata->headers=(WAVEHDR *)buffer;
	buffer+=(sizeof(WAVEHDR)*ga->numAudioBuffers);

	for(ui1=0;ui1<(unsigned int)ga->numAudioBuffers;ui1++) {

		#ifdef WIN_DEBUG
	    printf("\tPrep header %d/%d\n",ui1,ga->numAudioBuffers);
        #endif

	    pdata->headers[ui1].dwBufferLength=ga->audioBufferSize;
		pdata->WinBuffer[ui1]=pdata->headers[ui1].lpData=buffer;
		buffer+=ga->audioBufferSize;

		/* Prepare Headers */
		waveOutPrepareHeader(pdata->hWaveOut, &pdata->headers[ui1], sizeof(WAVEHDR));
		/* Pretend this new buffer was already played */
		pdata->headers[ui1].dwFlags|=WHDR_DONE;
	}

    #ifdef WIN_DEBUG
	printf("\tCreate Mutex\n");
    #endif

	/* Create Mutex */
	ga->buffer_mutex=CreateMutex(0, 0, NULL);
	if(ga->buffer_mutex==NULL) {
		ga->lastERR=ERR_MUTEX_INIT;
		goto err;
	}

	/* Init CriticalSection pointer */
	InitializeCriticalSection(&pdata->criticalSection);

	if (ga->max_voices) {
		#ifdef WIN_DEBUG
		printf("\tStart Playing Thread (bits=%d  channels=%d)\n",ga->bits, ga->channels);
		#endif

		/* Start playing thread */
		if(ga->bits==16) {
		if(ga->channels==1) thread_rtn= (void*) playthrd16m;
		else thread_rtn=(void *) playthrd16s;
		} else {
		if(ga->channels==1) thread_rtn= (void*) playthrd8m;
		else thread_rtn=(void *) playthrd8s;
		}

		ga->audio_thread=CreateThread(NULL,0, thread_rtn , ga, 0, NULL);
		if(ga->audio_thread==NULL) {
			ga->lastERR=ERR_THREAD_START;
			goto err;
		}
	}
    ga->audio_fd=1;
    return( ga->lastERR );

err:

    if(ga->samples!=NULL) free(ga->samples);
    if(ga->playing!=NULL) free(ga->playing);
    if(ga->platformData!=NULL) free(ga->platformData);
	if(pdata!=NULL && pdata->headers!=NULL) free(pdata->headers);
	if(pdata!=NULL && pdata->WinBuffer!=NULL) free(pdata->WinBuffer);
	if(ga->buffer_mutex!=NULL) CloseHandle(ga->buffer_mutex);
	ga->audio_fd = -1;
    return(ga->lastERR);
}

